__author__ = 'duarte'
from modules.parameters import ParameterSet, ParameterSpace, extract_nestvalid_dict
from modules.input_architect import EncodingLayer, InputNoise
from modules.net_architect import Network
from modules.io import set_storage_locations
from modules.signals import iterate_obj_list, SpikeList, empty
from modules.visualization import set_global_rcParams, InputPlots, plot_histogram, pretty_raster
from modules.analysis import characterize_population_activity
import cPickle as pickle
import numpy as np
import nest
import time
import matplotlib.pyplot as pl
import scipy.stats as st

# ######################################################################################################################
# Experiment options
# ======================================================================================================================
plot = True
display = True
save = True
debug = False

# ######################################################################################################################
# Extract parameters from file and build global ParameterSet
# ======================================================================================================================
# params_file = '../parameters/dc_noise_input.py'
params_file = '../parameters/spike_noise_input.py'

parameter_set = ParameterSpace(params_file)[0]
parameter_set = parameter_set.clean(termination='pars')

if not isinstance(parameter_set, ParameterSet):
	if isinstance(parameter_set, basestring) or isinstance(parameter_set, dict):
		parameter_set = ParameterSet(parameter_set)
	else:
		raise TypeError("parameter_set must be ParameterSet, string with full path to parameter file or dictionary")

# ######################################################################################################################
# Setup extra variables and parameters
# ======================================================================================================================
if plot:
	set_global_rcParams(parameter_set.kernel_pars['mpl_path'])
paths = set_storage_locations(parameter_set, save)

np.random.seed(parameter_set.kernel_pars['np_seed'])
results = dict()

# ######################################################################################################################
# Set kernel and simulation parameters
# ======================================================================================================================
print '\nRuning ParameterSet {0}'.format(parameter_set.label)
nest.ResetKernel()
nest.set_verbosity('M_WARNING')
nest.SetKernelStatus(extract_nestvalid_dict(parameter_set.kernel_pars.as_dict(), param_type='kernel'))

# ######################################################################################################################
# Build network
# ======================================================================================================================
net = Network(parameter_set.net_pars)

# ######################################################################################################################
# Randomize initial variable values
# ======================================================================================================================
for idx, n in enumerate(list(iterate_obj_list(net.populations))):
	if hasattr(parameter_set.net_pars, "randomize_neuron_pars"):
		randomize = parameter_set.net_pars.randomize_neuron_pars[idx]
		for k, v in randomize.items():
			n.randomize_initial_states(k, randomization_function=v[0], **v[1])

# ######################################################################################################################
# Build and connect input
# ======================================================================================================================
if hasattr(parameter_set, "input_pars"):
	# Current input (need to build noise signal)
	total_stimulation_time = parameter_set.kernel_pars.sim_time + parameter_set.kernel_pars.transient_t
	input_noise = InputNoise(parameter_set.input_pars.noise,
	                         stop_time=total_stimulation_time)
	input_noise.generate()
	input_noise.re_seed(parameter_set.kernel_pars.np_seed)

	if plot:
		inp_plot = InputPlots(stim_obj=None, input_obj=None, noise_obj=input_noise)
		inp_plot.plot_noise_component(display=display, save=False)

	enc_layer = EncodingLayer(parameter_set.encoding_pars, signal=input_noise)
	enc_layer.connect(parameter_set.encoding_pars, net)
else:
	# Poisson input
	enc_layer = EncodingLayer(parameter_set.encoding_pars)
	enc_layer.connect(parameter_set.encoding_pars, net)

# ######################################################################################################################
# Set-up Analysis
# ======================================================================================================================
net.connect_devices()

# ######################################################################################################################
# Connect Network
# ======================================================================================================================
net.connect_populations(parameter_set.connection_pars, progress=True)

# ######################################################################################################################
# Simulate
# ======================================================================================================================
if parameter_set.kernel_pars.transient_t:
	net.simulate(parameter_set.kernel_pars.transient_t)
	net.flush_records()

net.simulate(parameter_set.kernel_pars.sim_time)  # +.1 to acquire last step...

# ######################################################################################################################
# Extract and store data
# ======================================================================================================================
net.extract_population_activity()
net.extract_network_activity()
net.flush_records()

# ######################################################################################################################
# Analyse / plot data
# ======================================================================================================================
analysis_interval = [parameter_set.kernel_pars.transient_t,
	                     parameter_set.kernel_pars.sim_time + parameter_set.kernel_pars.transient_t]
parameter_set.analysis_pars.pop('label')
start_analysis = time.time()
results.update(characterize_population_activity(net, parameter_set, analysis_interval, epochs=None,
                                                color_map='jet', plot=plot,
                                                display=display, save=paths['figures']+paths['label'],
                                                color_subpop=True, analysis_pars=parameter_set.analysis_pars))
print "\nElapsed time (state characterization): {0}".format(str(time.time() - start_analysis))

if 'mean_I_ex' in results['analog_activity']['E'].keys():
	inh = np.array(results['analog_activity']['E']['mean_I_in'])
	exc = np.array(results['analog_activity']['E']['mean_I_ex'])
	ei_ratios = np.abs(np.abs(inh) - np.abs(exc))
	ei_ratios_corrected = np.abs(np.abs(inh - np.mean(inh)) - np.abs(exc - np.mean(exc)))
	print "EI amplitude difference: {0} +- {1}".format(str(np.mean(ei_ratios)), str(np.std(ei_ratios)))
	print "EI amplitude difference (amplitude corrected): {0} +- {1}".format(str(np.mean(ei_ratios_corrected)),
	                                                                         str(np.std(ei_ratios_corrected)))
	results['analog_activity']['E']['IE_ratio'] = np.mean(ei_ratios)
	results['analog_activity']['E']['IE_ratio_corrected'] = np.mean(ei_ratios_corrected)

# ######################################################################################################################
# Save data
# ======================================================================================================================
if save:
	with open(paths['results'] + 'Results_' + parameter_set.label, 'w') as f:
		pickle.dump(results, f)
	parameter_set.save(paths['parameters'] + 'Parameters_' + parameter_set.label)


# ####
# illustration
colors = [(0., 0., 1.), (1., 0., 0.)]

# def plot_distribution(data, axis, color, **kwargs):
# 	n, bins = plot_histogram(data, 100, ax=axis, color=color, mark_mean=True, **{'histtype':
# 		    'stepfilled', 'alpha': 0.6})
# 	approximate_pdf_isi = st.kde.gaussian_kde(data)
# 	x = np.linspace(np.min(data), np.max(data), 100)
# 	y = approximate_pdf_isi(x)
# 	y /= np.sum(y)
# 	axis.plot(x, y, color=color, lw=2)
# 	axis.set(**kwargs)
# 	axis.set_ylim([0., np.max(n)])
#
#
# expected_values = {'regularity': {'cvs': 1., 'lvs': 1., 'cvs_log': 0.25},
#                   'synchrony': {'ISI_distance': 0.5, 'SPIKE_distance': 0.3, 'SPIKE_sync_distance': 0.25}}
#
#
# fig = pl.figure()
# ax1 = fig.add_subplot(311)
# ax2 = fig.add_subplot(312)
# ax3 = fig.add_subplot(313)
# for idx_pop, n_pop in enumerate(net.populations):
# 	spk_list = n_pop.spiking_activity
# 	# plot_histogram(spk_list.mean_rates(), 100, ax=ax1, color=colors[idx_pop], mark_mean=True)
# 	plot_distribution(spk_list.mean_rates(), ax1, colors[idx_pop], **{'xlabel': 'Rate', 'ylabel': 'Freq.'})
# 	plot_distribution(spk_list.cv_isi(), ax2, colors[idx_pop], **{'xlabel': r'$\mathrm{CV_{ISI}}$', 'ylabel': 'Freq.'})
# 	plot_distribution(spk_list.pairwise_pearson_corrcoeff(500, time_bin=1., all_coef=True), ax3, colors[idx_pop],
# 	                  **{'xlabel': r'$\mathrm{CC}$', 'ylabel': 'Freq.'})
#
#
# 	plot_histogram(spk_list.cv_isi(), 100, ax=ax2, color=colors[idx_pop])
# 	plot_histogram(spk_list.pairwise_pearson_corrcoeff(500, time_bin=1., all_coef=True), 100)
# pl.show()

# plot EI current examples
fig = pl.figure(figsize=(8, 4))
ax4 = fig.add_subplot(211)
ax3 = fig.add_subplot(212)
name = 'E'
n_pop = net.populations[0]
pop_idx = parameter_set.net_pars.pop_names.index(name)
times = results['analog_activity'][name]['time_axis']
vm = results['analog_activity'][name]['single_Vm']
idx = results['analog_activity'][name]['single_idx']

if len(vm) != len(times):
	times = times[:-1]

ax4.plot(times, vm, 'k', lw=1)
idxs = vm.argsort()
if 'V_reset' in parameter_set.net_pars.neuron_pars[pop_idx].keys() and 'V_th' in parameter_set.net_pars.neuron_pars[
	pop_idx].keys():
	v_reset = parameter_set.net_pars.neuron_pars[pop_idx]['V_reset']
	v_th = parameter_set.net_pars.neuron_pars[pop_idx]['V_th']
	possible_spike_times = [t for t in idxs if (t < len(vm) - 1) and (vm[t + 1] == v_reset) and (vm[t] != v_reset)]
	ax4.vlines(times[possible_spike_times], v_th, 50., lw=1)
ax4.set_ylim(min(vm) - 5., 10.)
ax4.set_ylabel(r'$\mathrm{V_{m} [mV]}$')
ax4.set_xlim(min(times), max(times))
ax3.set_title('Neuron {0}'.format(str(idx)))
currents = [x for x in results['analog_activity'][name].keys() if x in ['I_in', 'I_ex']]
if not empty(currents):
	cl = ['r', 'b', 'gray']
	for iiddxx, nn_curr in enumerate(currents):
		ax3.plot(times, results['analog_activity'][name][nn_curr], c=cl[iiddxx], lw=1)
	ax3.set_xlim(min(times), max(times))
	ax3.set_ylabel(r'$\mathrm{I^{syn}} \mathrm{[nA]}$')
else:
	irrelevant_keys = ['single_Vm', 'single_idx']
	other_variables = [x for x in results['analog_activity'][name].keys() if x[:6] == 'single' and x
	                   not in irrelevant_keys]
	cl = ['g', 'k', 'gray']
	for iiddxx, nn_curr in enumerate(other_variables):
		ax3.plot(times, results['analog_activity'][name][nn_curr], c=cl[iiddxx], lw=1)
		ax3.set_xlim(min(times), max(times))
		ax3.set_ylabel('{0}'.format(nn_curr))
ax3.set_xlabel('Time [ms]')
pl.show()

# Illustrate activity
time_window = [parameter_set.kernel_pars.transient_t,
               parameter_set.kernel_pars.transient_t + parameter_set.kernel_pars.sim_time]  # [10000., 15000.]
select_N = [80, 20]
ctr = 0

fig = pl.figure(figsize=(8, 4))
# pl.axis('off')
fig.subplots_adjust(hspace=0)
ax1 = fig.add_subplot(111)
all_ids = []

for idx_pop, n_pop in enumerate(net.populations):
	spike_list = n_pop.spiking_activity.time_slice(t_start=time_window[0], t_stop=time_window[1])
	if not spike_list.empty():
		new_ids = np.intersect1d(spike_list.select_ids("cell.mean_rate() > 0"),
		                         spike_list.select_ids("cell.mean_rate() < 100"))
		chosen_ids = np.random.permutation(new_ids)[:int(select_N[idx_pop])]
		new_list = spike_list.id_slice(chosen_ids)

		neuron_ids = [np.where(x == np.unique(new_list.raw_data()[:, 1]))[0][0] for x in new_list.raw_data()[:, 1]]
		tmp = [(neuron_ids[idx], time) for idx, time in enumerate(new_list.raw_data()[:, 0])]
		new_spike_list = SpikeList(tmp, id_list=np.unique(neuron_ids), t_start=time_window[0], t_stop=time_window[1])

		# for idx, n in enumerate(tmp):
		# 	neuron_id = n[0] + ctr
			# ax1.vlines(n[1] - time_window[0], neuron_id - 0.4, neuron_id + 0.4, **{'color': colors[idx_pop], 'lw': 2.})
		ax1.plot([n[1] - time_window[0] for n in tmp], [n[0]+ctr for n in tmp], '.', **{'color': colors[idx_pop],
		                                                                                'ms': 5.})
		all_ids.append([n[0] + ctr for n in tmp])
		ctr += max(new_spike_list.id_list) + 1
		# ax.set_ylim(-0.5, select_N[idx_pop] - 0.5)
		# all_ids.append([n[0]+ctr for n in tmp])

idds = list(itertools.chain(*all_ids))
ax1.set_xlim(0., time_window[1] - time_window[0])
ax1.grid(False)
ax1.set_xlabel('Time [ms]')
ax1.set_yticks([])
ax1.set_ylim([min(idds), max(idds)])
pl.show()
